@@ -1,27 +1,27 @@
 from __future__ import annotations
 
 import functools
-from collections import deque
+from collections import defaultdict, deque
 from contextvars import ContextVar
 from typing import (
     Any,
-    Callable,
     ContextManager,
-    Coroutine,
+    DefaultDict,
     Deque,
     Dict,
-    Iterable,
     List,
+    Mapping,
     Optional,
+    Tuple,
     Union,
     cast,
 )
 
+from di._dag import topsort
 from di._inspect import is_async_gen_callable, is_coroutine_callable
 from di._local_scope_context import LocalScopeContext
 from di._state import ContainerState
 from di._task import AsyncTask, SyncTask, Task
-from di._topsort import topsort
 from di.exceptions import (
     DependencyRegistryError,
     DuplicateScopeError,
@@ -31,11 +31,18 @@ from di.exceptions import (
 from di.executors import DefaultExecutor
 from di.types import FusedContextManager
 from di.types.dependencies import DependantProtocol, DependencyParameter
-from di.types.executor import AsyncExecutor, SyncExecutor, Values
-from di.types.providers import Dependency, DependencyProviderType, DependencyType
+from di.types.executor import AsyncExecutor, SyncExecutor
+from di.types.executor import Task as ExecutorTask
+from di.types.providers import (
+    DependencyProvider,
+    DependencyProviderType,
+    DependencyType,
+)
 from di.types.scopes import Scope
 from di.types.solved import SolvedDependency
 
+Dependency = Any
+
 
 class Container:
     _context: ContextVar[ContainerState]
@@ -109,7 +116,6 @@ class Container:
     def solve(
         self,
         dependency: DependantProtocol[DependencyType],
-        siblings: Optional[Iterable[DependantProtocol[Any]]] = None,
     ) -> SolvedDependency[DependencyType]:
         """Solve a dependency.
 
@@ -120,7 +126,7 @@ class Container:
         """
 
         if dependency.call in self._state.binds:
-            dependency = self._state.binds[dependency.call]  # type: ignore
+            dependency = self._state.binds[dependency.call]
 
         param_graph: Dict[
             DependantProtocol[Any],
@@ -172,33 +178,42 @@ class Container:
                     dep_dag[dep].append(subdep)
                     if subdep not in dep_registry:
                         q.append(subdep)
+        tasks = self._build_tasks(param_graph, topsort(dep_dag))
+        dependency_dag = {dep: set(subdeps) for dep, subdeps in dep_dag.items()}
+        # we store several things:
+        # - the dependency itself
+        # - dag, which includes parameter information and is public
+        # - _dependency_dag: same as dag but without param info and w/o repeated params (so a set instead of a list)
+        # - _tasks: a mapping of dependency to tasks
+        return SolvedDependency(
+            dependency=dependency,
+            dag=param_graph,
+            _dependency_dag=dependency_dag,
+            _tasks=tasks,
+        )
 
-        siblings = siblings or ()
-        for sibling in siblings:
-            solved = self.solve(sibling)
-            param_graph.update(solved.dag)
-            dep_dag.update(
-                {
-                    dep: [p.dependency for p in params]
-                    for dep, params in solved.dag.items()
-                }
-            )
-
-        topsorted_groups = topsort(dep_dag)
-        tasks = self._build_tasks(topsorted_groups, param_graph)
-        return SolvedDependency(dependency=dependency, dag=param_graph, _tasks=tasks)  # type: ignore
-
-    def _build_task(
+    def _build_tasks(
         self,
-        dependency: DependantProtocol[DependencyType],
-        tasks: Dict[
-            DependantProtocol[Any], Union[AsyncTask[Dependency], SyncTask[Dependency]]
-        ],
         dag: Dict[
             DependantProtocol[Any],
             List[DependencyParameter[DependantProtocol[Any]]],
         ],
-    ) -> Union[AsyncTask[DependencyType], SyncTask[DependencyType]]:
+        topsorted: List[DependantProtocol[Any]],
+    ) -> Dict[DependantProtocol[Any], Union[AsyncTask[Any], SyncTask[Any]]]:
+        tasks: Dict[DependantProtocol[Any], Union[AsyncTask[Any], SyncTask[Any]]] = {}
+        for dep in reversed(topsorted):
+            tasks[dep] = self._build_task(dep, tasks, dag)
+        return tasks
+
+    def _build_task(
+        self,
+        dependency: DependantProtocol[Any],
+        tasks: Dict[DependantProtocol[Any], Union[AsyncTask[Any], SyncTask[Any]]],
+        dag: Dict[
+            DependantProtocol[Any],
+            List[DependencyParameter[DependantProtocol[Any]]],
+        ],
+    ) -> Union[AsyncTask[Any], SyncTask[Any]]:
 
         task_dependencies: List[DependencyParameter[Task[Any]]] = [
             DependencyParameter(
@@ -214,23 +229,6 @@ class Container:
         else:
             return SyncTask(dependant=dependency, dependencies=task_dependencies)
 
-    def _build_tasks(
-        self,
-        topsort: List[List[DependantProtocol[Any]]],
-        dag: Dict[
-            DependantProtocol[Any],
-            List[DependencyParameter[DependantProtocol[Any]]],
-        ],
-    ) -> List[List[Union[AsyncTask[Dependency], SyncTask[Dependency]]]]:
-        tasks: Dict[
-            DependantProtocol[Any], Union[AsyncTask[Dependency], SyncTask[Dependency]]
-        ] = {}
-        for group in reversed(topsort):
-            for dep in group:
-                if dep not in tasks:
-                    tasks[dep] = self._build_task(dep, tasks, dag)
-        return list(reversed([[tasks[dep] for dep in group] for group in topsort]))
-
     def _validate_scopes(self, solved: SolvedDependency[Dependency]) -> None:
         """Validate that dependencies all have a valid scope and
         that dependencies only depend on outer scopes or their own scope.
@@ -263,68 +261,111 @@ class Container:
                 check_scope(subdep)
                 check_is_inner(dep, subdep)
 
+    def _prepare_execution(
+        self,
+        solved: SolvedDependency[Any],
+        *,
+        validate_scopes: bool = True,
+        values: Optional[Mapping[DependencyProvider, Any]] = None,
+    ) -> Tuple[Dict[DependantProtocol[Any], Any], List[ExecutorTask]]:
+        results: Dict[DependantProtocol[Any], Any] = {}
+        values = values or {}
+        if validate_scopes:
+            self._validate_scopes(solved)
+
+        tasks = solved._tasks
+
+        # Build a DAG of Tasks that we actually need to execute
+        # this allows us to prune subtrees that will come from cached values
+        unvisited = deque([solved.dependency])
+        # Make DAG values a set to account for dependencies that depend on the the same
+        # sub dependency in more than one param (`def func(a: A, a_again: A)`)
+        dag: Dict[DependantProtocol[Any], List[DependantProtocol[Any]]] = {}
+        dependant_dag: DefaultDict[
+            DependantProtocol[Any], List[Task[Any]]
+        ] = defaultdict(list)
+        while unvisited:
+            dep = unvisited.pop()
+            if dep in dag:
+                continue
+            task = tasks[dep]
+            # task the dependency is cached or was provided by value
+            # we don't need to compute it or any of it's dependencies
+            if not task.from_cache_or_values(self._state, results, values):
+                # otherwise, we add it to our DAG and visit it's children
+                dag[dep] = []
+                for subdep in solved._dependency_dag[dep]:
+                    if not tasks[subdep].from_cache_or_values(
+                        self._state, results, values
+                    ):
+                        dependant_dag[subdep].append(tasks[dep])
+                        dag[dep].append(subdep)
+                        if subdep not in dag:
+                            unvisited.append(subdep)
+
+        dependency_counts = {dep: len(subdeps) for dep, subdeps in dag.items()}
+
+        leafs: List[ExecutorTask] = []
+        for dep, depcount in dependency_counts.items():
+            if depcount == 0:
+                leaf = functools.partial(
+                    tasks[dep].compute,
+                    state=self._state,
+                    results=results,
+                    values=values,
+                    dependency_counts=dependency_counts,
+                    dependants=dependant_dag,
+                )
+                leafs.append(leaf)  # type: ignore[arg-type] # beacuse of partial
+
+        return results, leafs
+
     def execute_sync(
         self,
         solved: SolvedDependency[DependencyType],
         *,
         validate_scopes: bool = True,
-        values: Optional[Values] = None,
+        values: Optional[Mapping[DependencyProvider, Any]] = None,
     ) -> DependencyType:
         """Execute an already solved dependency.
 
         If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
         and then disable scope validation in subsequent runs with `validate_scope=False`.
         """
-        results: Dict[Task[Any], Any] = {}
-        values = values or {}
         with self.enter_local_scope(self._default_scope):
-            if validate_scopes:
-                self._validate_scopes(solved)
-
-            tasks = solved._tasks  # type: ignore
-
+            results, queue = self._prepare_execution(
+                solved, validate_scopes=validate_scopes, values=values
+            )
             if not hasattr(self._executor, "execute_sync"):
                 raise TypeError(
                     "execute_sync requires an executor implementing the SyncExecutor protocol"
                 )
             executor = cast(SyncExecutor, self._executor)
-
-            stateful_tasks = [
-                [functools.partial(t.compute, self._state, results) for t in group]
-                for group in tasks
-            ]
-            return executor.execute_sync(stateful_tasks, lambda: results[tasks[-1][0]], values)  # type: ignore
+            if queue:
+                executor.execute_sync(queue)
+            return results[solved.dependency]
 
     async def execute_async(
         self,
         solved: SolvedDependency[DependencyType],
         *,
         validate_scopes: bool = True,
-        values: Optional[Values] = None,
+        values: Optional[Mapping[DependencyProvider, Any]] = None,
     ) -> DependencyType:
         """Execute an already solved dependency.
 
         If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
         and then disable scope validation in subsequent runs with `validate_scope=False`.
         """
-        results: Dict[Task[Any], Any] = {}
-        values = values or {}
         async with self.enter_local_scope(self._default_scope):
-            if validate_scopes:
-                self._validate_scopes(solved)
-
-            tasks = solved._tasks  # type: ignore
-
+            results, queue = self._prepare_execution(
+                solved, validate_scopes=validate_scopes, values=values
+            )
             if not hasattr(self._executor, "execute_async"):
                 raise TypeError(
                     "execute_async requires an executor implementing the AsyncExecutor protocol"
                 )
             executor = cast(AsyncExecutor, self._executor)
-
-            stateful_tasks: List[
-                List[Callable[[], Union[Coroutine[Any, Any, None], None]]]
-            ] = [
-                [functools.partial(t.compute, self._state, results) for t in group]
-                for group in tasks
-            ]
-            return await executor.execute_async(stateful_tasks, lambda: results[tasks[-1][0]], values)  # type: ignore
+            if queue:
+                await executor.execute_async(queue)
+            return results[solved.dependency]
